@@ -4,7 +4,7 @@
 package POE::Resource::Signals;
 
 use vars qw($VERSION);
-$VERSION = '1.294'; # NOTE - Should be #.### (three decimal places)
+$VERSION = '1.367'; # NOTE - Should be #.### (three decimal places)
 
 # These methods are folded into POE::Kernel;
 package POE::Kernel;
@@ -18,17 +18,29 @@ use POSIX qw(:sys_wait_h sigprocmask SIG_SETMASK);
 ### Map watched signal names to the sessions that are watching them
 ### and the events that must be delivered when they occur.
 
+sub SEV_EVENT   () { 0 }
+sub SEV_ARGS    () { 1 }
+sub SEV_SESSION () { 2 }
+
 my %kr_signals;
 #  ( $signal_name =>
-#    { $session_reference => [ $event_name, $event_args, ],
+#    { $session_id =>
+#     [ $event_name,    SEV_EVENT
+#       $event_args,    SEV_ARGS
+#       $session_ref,   SEV_SESSION
+#     ],
 #      ...,
 #    },
 #    ...,
 #  );
 
 my %kr_sessions_to_signals;
-#  ( $session =>
-#    { $signal_name => [ $event_name, $event_args ],
+#  ( $session_id =>
+#    { $signal_name =>
+#      [ $event_name,   SEV_EVENT
+#        $event_args,   SEV_ARGS
+#        $session_ref,  SEV_SESSION
+#      ],
 #      ...,
 #    },
 #    ...,
@@ -36,7 +48,7 @@ my %kr_sessions_to_signals;
 
 my %kr_pids_to_events;
 # { $pid =>
-#   { $session =>
+#   { $session_id =>
 #     [ $blessed_session,   # PID_SESSION
 #       $event_name,        # PID_EVENT
 #       $args,              # PID_ARGS
@@ -45,14 +57,35 @@ my %kr_pids_to_events;
 # }
 
 my %kr_sessions_to_pids;
-# { $session => { $pid => 1 } }
+# { $session_id => { $pid => 1 } }
 
 sub PID_SESSION () { 0 }
 sub PID_EVENT   () { 1 }
 sub PID_ARGS    () { 2 }
 
+sub _data_sig_relocate_kernel_id {
+  my ($self, $old_id, $new_id) = @_;
+
+  while (my ($signal, $sig_rec) = each %kr_signals) {
+    next unless exists $sig_rec->{$old_id};
+    $sig_rec->{$new_id} = delete $sig_rec->{$old_id};
+  }
+
+  $kr_sessions_to_signals{$new_id} = delete $kr_sessions_to_signals{$old_id}
+    if exists $kr_sessions_to_signals{$old_id};
+
+  while (my ($pid, $pid_rec) = each %kr_pids_to_events) {
+    next unless exists $pid_rec->{$old_id};
+    $pid_rec->{$new_id} = delete $pid_rec->{$old_id};
+  }
+
+  $kr_sessions_to_pids{$new_id} = delete $kr_sessions_to_pids{$old_id}
+    if exists $kr_sessions_to_pids{$old_id};
+}
+
 # Bookkeeping per dispatched signal.
 
+# TODO - Why not lexicals?
 use vars (
  '@kr_signaled_sessions',            # The sessions touched by a signal.
  '$kr_signal_total_handled',         # How many sessions handled a signal.
@@ -68,8 +101,14 @@ use vars (
 # already been queued.
 my $polling_for_signals = 0;
 
-# A flag determining whether there are child processes.
-my $kr_child_procs = exists($INC{'Apache.pm'}) ? 0 : ( USE_SIGCHLD ? 0 : 1 );
+# There may be latent subprocesses in some environments.
+# Or we may need to "always loop once" if we're polling for SIGCHLD.
+# This constant lets us define those exceptional cases.
+# We had some in the past, but as of 2013-10-06 we seem to have
+# eliminated those special cases.
+use constant BASE_SIGCHLD_COUNT => 0;
+
+my $kr_has_child_procs = BASE_SIGCHLD_COUNT;
 
 # A list of special signal types.  Signals that aren't listed here are
 # benign (they do not kill sessions at all).  "Terminal" signals are
@@ -162,7 +201,7 @@ sub _data_sig_reset_procs {
   # least once.  Starts false when running in an Apache handler so our
   # SIGCHLD hijinks don't interfere with the web server.
   $self->_data_sig_cease_polling();
-  $kr_child_procs = exists($INC{'Apache.pm'}) ? 0 : ( USE_SIGCHLD ? 0 : 1 );
+  $kr_has_child_procs = BASE_SIGCHLD_COUNT;
 }
 
 
@@ -186,35 +225,51 @@ sub _data_sig_finalize {
   while (my ($sig, $sig_rec) = each(%kr_signals)) {
     $finalized_ok = 0;
     _warn "!!! Leaked signal $sig\n";
-    while (my ($ses, $ses_rec) = each(%{$kr_signals{$sig}})) {
-      my ($event, $args) = @$ses_rec;
-      _warn "!!!\t$ses = $event (@$args)\n";
+    while (my ($sid, $ses_rec) = each(%{$kr_signals{$sig}})) {
+      my ($event, $args, $session) = @$ses_rec;
+      _warn "!!!\t$sid = $session -> $event (@$args)\n";
     }
   }
 
-  while (my ($ses, $ses_rec) = each(%kr_sessions_to_signals)) {
+  while (my ($sid, $ses_rec) = each(%kr_sessions_to_signals)) {
     $finalized_ok = 0;
-    _warn "!!! Leaked signal cross-reference: $ses\n";
-    while (my ($sig, $sig_rec) = each(%{$kr_signals{$ses}})) {
+    _warn "!!! Leaked signal cross-reference: $sid\n";
+    while (my ($sig, $sig_rec) = each(%{$kr_signals{$sid}})) {
       my ($event, $args) = @$sig_rec;
       _warn "!!!\t$sig = $event (@$args)\n";
     }
   }
 
-  while (my ($ses, $pid_rec) = each(%kr_sessions_to_pids)) {
+  while (my ($sid, $pid_rec) = each(%kr_sessions_to_pids)) {
     $finalized_ok = 0;
     my @pids = keys %$pid_rec;
-    _warn "!!! Leaked session to PID map: $ses -> (@pids)\n";
+    _warn "!!! Leaked session to PID map: $sid -> (@pids)\n";
   }
 
   while (my ($pid, $ses_rec) = each(%kr_pids_to_events)) {
     $finalized_ok = 0;
     _warn "!!! Leaked PID to event map: $pid\n";
-    while (my ($ses, $ev_rec) = each %$ses_rec) {
+    while (my ($sid, $ev_rec, $ses) = each %$ses_rec) {
       _warn "!!!\t$ses -> $ev_rec->[PID_EVENT] (@{$ev_rec->[PID_ARGS]})\n";
     }
   }
 
+  if ($kr_has_child_procs) {
+    _warn "!!! Kernel has $kr_has_child_procs child process(es).\n";
+  }
+
+  if ($polling_for_signals) {
+    _warn "!!! Finalizing signals while polling is active.\n";
+  }
+
+  if (USE_SIGNAL_PIPE and $self->_data_sig_pipe_has_signals()) {
+    _warn "!!! Finalizing signals while signal pipe contains messages.\n";
+  }
+
+  if (exists $kr_signals{CHLD}) {
+    _warn "!!! Finalizing signals while a blanket _child signal is watched.\n";
+  }
+
   %_safe_signals = ();
 
   unless (RUNNING_IN_HELL) {
@@ -223,18 +278,28 @@ sub _data_sig_finalize {
 
     my $leaked_children = 0;
 
-    until ((my $pid = waitpid( -1, 0 )) == -1) {
-      _warn( "!!! Child process PID:$pid reaped: $!\n" ) if $pid;
+    PROCESS: until ((my $pid = waitpid( -1, WNOHANG )) == -1) {
       $finalized_ok = 0;
       $leaked_children++;
+
+      if ($pid == 0) {
+        _warn(
+          "!!! At least one child process is still running " .
+          "when POE::Kernel->run() is ready to return.\n"
+        );
+        last PROCESS;
+      }
+
+      _warn(
+        "!!! Stopped child process (PID $pid) reaped " .
+          "when POE::Kernel->run() is ready to return.\n"
+      );
     }
 
     if ($leaked_children) {
-      _warn(
-        "!!! Your program may not be using sig_child() to reap processes.\n",
-        "!!! In extreme cases, your program can force a system reboot\n",
-        "!!! if this resource leakage is not corrected.\n",
-      );
+      _warn("!!! Be sure to use sig_child() to reap child processes.\n");
+      _warn("!!! In extreme cases, failure to reap child processes has\n");
+      _warn("!!! resulted in a slow 'fork bomb' that has halted systems.\n");
     }
   }
 
@@ -246,13 +311,16 @@ sub _data_sig_finalize {
 sub _data_sig_add {
   my ($self, $session, $signal, $event, $args) = @_;
 
-  $kr_sessions_to_signals{$session}->{$signal} = [ $event, $args || [] ];
-  $self->_data_sig_signal_watch($session, $signal);
-  $kr_signals{$signal}->{$session} = [ $event, $args || [] ];
+  my $sid = $session->ID;
+  $kr_sessions_to_signals{$sid}->{$signal} = [ $event, $args || [], $session ];
+  $self->_data_sig_signal_watch($sid, $signal);
+  $kr_signals{$signal}->{$sid} = [ $event, $args || [], $session ];
 }
 
 sub _data_sig_signal_watch {
-  my ($self, $session, $signal) = @_;
+  my ($self, $sid, $signal) = @_;
+
+  # TODO - $sid not used?
 
   # First session to watch the signal.
   # Ask the event loop to watch the signal.
@@ -266,7 +334,9 @@ sub _data_sig_signal_watch {
 }
 
 sub _data_sig_signal_ignore {
-  my ($self, $session, $signal) = @_;
+  my ($self, $sid, $signal) = @_;
+
+  # TODO - $sid not used?
 
   if (
     !exists($kr_signals{$signal}) and
@@ -280,18 +350,18 @@ sub _data_sig_signal_ignore {
 ### Remove a signal from a session.
 
 sub _data_sig_remove {
-  my ($self, $session, $signal) = @_;
+  my ($self, $sid, $signal) = @_;
 
-  delete $kr_sessions_to_signals{$session}->{$signal};
-  delete $kr_sessions_to_signals{$session}
-    unless keys(%{$kr_sessions_to_signals{$session}});
+  delete $kr_sessions_to_signals{$sid}->{$signal};
+  delete $kr_sessions_to_signals{$sid}
+    unless keys(%{$kr_sessions_to_signals{$sid}});
 
-  delete $kr_signals{$signal}->{$session};
+  delete $kr_signals{$signal}->{$sid};
 
   # Last watcher for that signal.  Stop watching it internally.
   unless (keys %{$kr_signals{$signal}}) {
     delete $kr_signals{$signal};
-    $self->_data_sig_signal_ignore($session, $signal);
+    $self->_data_sig_signal_ignore($sid, $signal);
   }
 }
 
@@ -302,17 +372,17 @@ sub _data_sig_remove {
 # mean that the session really doesn't exist.  Should we care?
 
 sub _data_sig_clear_session {
-  my ($self, $session) = @_;
+  my ($self, $sid) = @_;
 
-  if (exists $kr_sessions_to_signals{$session}) { # avoid autoviv
-    foreach (keys %{$kr_sessions_to_signals{$session}}) {
-      $self->_data_sig_remove($session, $_);
+  if (exists $kr_sessions_to_signals{$sid}) { # avoid autoviv
+    foreach (keys %{$kr_sessions_to_signals{$sid}}) {
+      $self->_data_sig_remove($sid, $_);
     }
   }
 
-  if (exists $kr_sessions_to_pids{$session}) { # avoid autoviv
-    foreach (keys %{$kr_sessions_to_pids{$session}}) {
-      $self->_data_sig_pid_ignore($session, $_);
+  if (exists $kr_sessions_to_pids{$sid}) { # avoid autoviv
+    foreach (keys %{$kr_sessions_to_pids{$sid}}) {
+      $self->_data_sig_pid_ignore($sid, $_);
     }
   }
 }
@@ -322,54 +392,68 @@ sub _data_sig_clear_session {
 sub _data_sig_pid_watch {
   my ($self, $session, $pid, $event, $args) = @_;
 
-  $kr_pids_to_events{$pid}{$session} = [
+  my $sid = $session->ID;
+
+  $kr_pids_to_events{$pid}{$sid} = [
     $session, # PID_SESSION
     $event,   # PID_EVENT
     $args,    # PID_ARGS
   ];
 
-  $self->_data_sig_signal_watch($session, "CHLD");
+  $self->_data_sig_signal_watch($sid, "CHLD");
 
-  $kr_sessions_to_pids{$session}{$pid} = 1;
-  $self->_data_ses_refcount_inc($session);
+  $kr_sessions_to_pids{$sid}{$pid} = 1;
+  $self->_data_ses_refcount_inc($sid);
 
   # Assume there's a child process.  This will be corrected on the
   # next polling interval.
-  $kr_child_procs++ unless USE_SIGCHLD;
+  $kr_has_child_procs++ unless USE_SIGCHLD;
 }
 
 sub _data_sig_pid_ignore {
-  my ($self, $session, $pid) = @_;
+  my ($self, $sid, $pid) = @_;
 
   # Remove PID to event mapping.
 
-  delete $kr_pids_to_events{$pid}{$session};
+  delete $kr_pids_to_events{$pid}{$sid};
   delete $kr_pids_to_events{$pid} unless (
     keys %{$kr_pids_to_events{$pid}}
   );
 
   # Remove session to PID mapping.
 
-  delete $kr_sessions_to_pids{$session}{$pid};
-  unless (keys %{$kr_sessions_to_pids{$session}}) {
-    delete $kr_sessions_to_pids{$session};
-    $self->_data_sig_signal_ignore($session, "CHLD");
+  delete $kr_sessions_to_pids{$sid}{$pid};
+  unless (keys %{$kr_sessions_to_pids{$sid}}) {
+    delete $kr_sessions_to_pids{$sid};
+    $self->_data_sig_signal_ignore($sid, "CHLD");
   }
 
-  $self->_data_ses_refcount_dec($session);
+  $self->_data_ses_refcount_dec($sid);
 }
 
-sub _data_sig_pids_ses {
-  my ($self, $session) = @_;
-  return 0 unless exists $kr_sessions_to_pids{$session};
-  return scalar keys %{$kr_sessions_to_pids{$session}};
+sub _data_sig_session_awaits_pids {
+  my ($self, $sid) = @_;
+
+  # There must be child processes or pending signals.
+  # Watching PIDs doesn't matter if there are none to be reaped.
+  return 0 unless $kr_has_child_procs or $self->_data_sig_pipe_has_signals();
+
+  # This session is watching at least one PID with sig_child().
+  # TODO - Watching a non-existent PID is legal but ill-advised.
+  return 1 if exists $kr_sessions_to_pids{$sid};
+
+  # Is the session waiting for a blanket sig(CHLD)?
+  return(
+    (exists $kr_sessions_to_signals{$sid}) &&
+    (exists $kr_sessions_to_signals{$sid}{CHLD})
+  );
 }
 
 sub _data_sig_pids_is_ses_watching {
-  my ($self, $session, $pid) = @_;
+  my ($self, $sid, $pid) = @_;
   return(
-    exists($kr_sessions_to_pids{$session}) &&
-    exists($kr_sessions_to_pids{$session}{$pid})
+    exists($kr_sessions_to_pids{$sid}) &&
+    exists($kr_sessions_to_pids{$sid}{$pid})
   );
 }
 
@@ -408,8 +492,9 @@ sub _data_sig_explicitly_watched {
 ### for introspection.
 
 sub _data_sig_watched_by_session {
-  my ($self, $session) = @_;
-  return %{$kr_sessions_to_signals{$session}};
+  my ($self, $sid) = @_;
+  return unless exists $kr_sessions_to_signals{$sid};
+  return %{$kr_sessions_to_signals{$sid}};
 }
 
 ### Which sessions are watching a signal?
@@ -435,10 +520,10 @@ sub _data_sig_handled_status {
 ### the shorter one.
 
 sub _data_sig_is_watched_by_session {
-  my ($self, $signal, $session) = @_;
+  my ($self, $signal, $sid) = @_;
   return(
     exists($kr_signals{$signal}) &&
-    exists($kr_signals{$signal}->{$session})
+    exists($kr_signals{$signal}->{$sid})
   );
 }
 
@@ -454,21 +539,22 @@ sub _data_sig_free_terminated_sessions {
     ($kr_signal_type & SIGTYPE_TERMINAL and !$kr_signal_total_handled)
   ) {
     foreach my $dead_session (@kr_signaled_sessions) {
-      next unless $self->_data_ses_exists($dead_session);
+      next unless $self->_data_ses_exists($dead_session->ID);
+
       if (TRACE_SIGNALS) {
         _warn(
           "<sg> stopping signaled session ",
-          $self->_data_alias_loggable($dead_session)
+          $self->_data_alias_loggable($dead_session->ID)
         );
       }
 
-      $self->_data_ses_stop($dead_session);
+      $self->_data_ses_stop($dead_session->ID);
     }
   }
 
   # Erase @kr_signaled_sessions, or they will leak until the next
   # signal.
-  undef @kr_signaled_sessions;
+  @kr_signaled_sessions = ();
 }
 
 ### A signal has touched a session.  Record this fact for later
@@ -504,7 +590,7 @@ sub _data_sig_enqueue_poll_event {
 
     $self->_data_ev_enqueue(
       $self, $self, EN_SCPOLL, ET_SCPOLL, [ $signal ],
-      __FILE__, __LINE__, undef, time(),
+      __FILE__, __LINE__, undef
     );
   } else {
     return if $self->_data_ses_count() < 1;
@@ -512,7 +598,7 @@ sub _data_sig_enqueue_poll_event {
 
     $self->_data_ev_enqueue(
       $self, $self, EN_SCPOLL, ET_SCPOLL, [ $signal ],
-      __FILE__, __LINE__, undef, time() + POE::Kernel::CHILD_POLLING_INTERVAL(),
+      __FILE__, __LINE__, undef, walltime(), POE::Kernel::CHILD_POLLING_INTERVAL(),
     );
   }
 }
@@ -526,11 +612,45 @@ sub _data_sig_handle_poll_event {
 
   if (TRACE_SIGNALS) {
     _warn(
-      "<sg> POE::Kernel is polling for signals at " . time() .
+      "<sg> POE::Kernel is polling for signals at " . monotime() .
       (USE_SIGCHLD ? " due to SIGCHLD" : "")
     );
   }
 
+  $self->_data_sig_reap_pids();
+
+  # The poll loop is over.  Resume slowly polling for signals.
+
+  if (USE_SIGCHLD) {
+    if (TRACE_SIGNALS) {
+      _warn("<sg> POE::Kernel has reset the SIG$signal handler");
+    }
+    # Per https://rt.cpan.org/Ticket/Display.html?id=45109 setting the
+    # signal handler must be done after reaping the outstanding child
+    # processes, at least on SysV systems like HP-UX.
+    $SIG{$signal} = \&_loop_signal_handler_chld;
+  }
+  else {
+    # The poll loop is over.  Resume slowly polling for signals.
+
+    if ($polling_for_signals) {
+      if (TRACE_SIGNALS) {
+        _warn("<sg> POE::Kernel will poll again after a delay");
+      }
+      $self->_data_sig_enqueue_poll_event($signal);
+    }
+    else {
+      if (TRACE_SIGNALS) {
+        _warn("<sg> POE::Kernel SIGCHLD poll loop paused");
+      }
+      $self->_idle_queue_shrink();
+    }
+  }
+}
+
+sub _data_sig_reap_pids {
+  my $self = shift();
+
   # Reap children for as long as waitpid(2) says something
   # interesting has happened.
   # TODO This has a possibility of an infinite loop, but so far it
@@ -541,7 +661,7 @@ sub _data_sig_handle_poll_event {
     # waitpid(2) returned a process ID.  Emit an appropriate SIGCHLD
     # event and loop around again.
 
-    if ((RUNNING_IN_HELL and $pid < -1) or ($pid > 0)) {
+    if (($pid > 0) or (RUNNING_IN_HELL and $pid < -1)) {
       if (RUNNING_IN_HELL or WIFEXITED($?) or WIFSIGNALED($?)) {
 
         if (TRACE_SIGNALS) {
@@ -553,13 +673,13 @@ sub _data_sig_handle_poll_event {
 
         if (exists $kr_pids_to_events{$pid}) {
           my @sessions_to_clear;
-          while (my ($ses_key, $ses_rec) = each %{$kr_pids_to_events{$pid}}) {
+          while (my ($sid, $ses_rec) = each %{$kr_pids_to_events{$pid}}) {
             $self->_data_ev_enqueue(
               $ses_rec->[PID_SESSION], $self, $ses_rec->[PID_EVENT], ET_SIGCLD,
               [ 'CHLD', $pid, $?, @{$ses_rec->[PID_ARGS]} ],
-              __FILE__, __LINE__, undef, time(),
+              __FILE__, __LINE__, undef
             );
-            push @sessions_to_clear, $ses_rec->[PID_SESSION];
+            push @sessions_to_clear, $sid;
           }
           $self->_data_sig_pid_ignore($_, $pid) foreach @sessions_to_clear;
         }
@@ -567,7 +687,7 @@ sub _data_sig_handle_poll_event {
         # Kick off a SIGCHLD cascade.
         $self->_data_ev_enqueue(
           $self, $self, EN_SIGNAL, ET_SIGNAL, [ 'CHLD', $pid, $? ],
-          __FILE__, __LINE__, undef, time(),
+          __FILE__, __LINE__, undef
         );
       }
       elsif (TRACE_SIGNALS) {
@@ -587,8 +707,7 @@ sub _data_sig_handle_poll_event {
     #
     # TODO - Find a way to test this.
 
-    _trap "internal consistency error: waitpid returned $pid"
-    if $pid != -1;
+    _trap "internal consistency error: waitpid returned $pid" if $pid != -1;
 
     # If the error is an interrupted syscall, poll again right away.
 
@@ -622,44 +741,28 @@ sub _data_sig_handle_poll_event {
     last;
   }
 
-  # If waitpid() returned 0, then we have child processes.
+  # Remember whether there are more processes to reap.
 
-  $kr_child_procs = !$pid;
-
-  if (USE_SIGCHLD) {
-    if (TRACE_SIGNALS) {
-      _warn("<sg> POE::Kernel has reset the SIG$signal handler");
-    }
-    # Per https://rt.cpan.org/Ticket/Display.html?id=45109 setting the
-    # signal handler must be done after reaping the outstanding child
-    # processes, at least on SysV systems like HP-UX.
-    $SIG{$signal} = \&_loop_signal_handler_chld;
-  }
-  else {
-    # The poll loop is over.  Resume slowly polling for signals.
-
-    if ($polling_for_signals) {
-      if (TRACE_SIGNALS) {
-        _warn("<sg> POE::Kernel will poll again after a delay");
-      }
-      $self->_data_sig_enqueue_poll_event($signal);
-    }
-    else {
-      if (TRACE_SIGNALS) {
-        _warn("<sg> POE::Kernel SIGCHLD poll loop paused");
-      }
-      $self->_idle_queue_shrink();
-    }
-  }
+  $kr_has_child_procs = !$pid;
 }
 
 # Are there child processes worth waiting for?
 # We don't really care if we're not polling for signals.
-# TODO - Will this change?
 
-sub _data_sig_child_procs {
-  return if !USE_SIGCHLD and !$polling_for_signals;
-  return $kr_child_procs;
+sub _data_sig_kernel_awaits_pids {
+  my $self = shift();
+
+  return 0 if !USE_SIGCHLD and !$polling_for_signals;
+
+  # There must be child processes or pending signals.
+  return 0 unless $kr_has_child_procs or $self->_data_sig_pipe_has_signals();
+
+  # At least one session is watching an explicit PID.
+  # TODO - Watching a non-existent PID is legal but ill-advised.
+  return 1 if scalar keys %kr_pids_to_events;
+
+  # Is the session waiting for a blanket sig(CHLD)?
+  return exists $kr_signals{CHLD};
 }
 
 ######################
@@ -672,40 +775,43 @@ sub _data_sig_child_procs {
 ##  - a top handler, which sends the signal number over a one-way pipe.
 ##  - a bottom handler, which is called when this number is received in the
 ##  main loop.
+## The top handler will send a packet of PID and number.  We need the PID
+## because of the race condition with signals in perl; signals meant for the
+## parent end up in both the parent and child.  So we check the PID to make
+## sure it was intended for the child.  We use 'ii' (2 ints, aka 8 bytes)
+## and not 'iC' (int+byte, aka 5 bytes) because we want a small factor of
+## the buffer size in the hopes of never getting a short read.  Ever.
 
 use vars qw( $signal_pipe_read_fd );
-my( $signal_pipe_write, $signal_pipe_read, $signal_pipe_pid,
-    $signal_mask_none, $signal_mask_all, %SIG2NUM, %NUM2SIG );
+my(
+  $signal_pipe_write,
+  $signal_pipe_read,
+  $signal_pipe_pid,
+  $signal_mask_none,
+  $signal_mask_all,
+
+  @pending_signals,
+);
+
+sub SIGINFO_NAME    () { 0 }
+sub SIGINFO_SRC_PID () { 1 }
+
+
+sub _data_sig_pipe_has_signals {
+  my $self = shift();
+  return unless $signal_pipe_read;
+  my $vec = '';
+  vec($vec, fileno($signal_pipe_read), 1) = 1;
+
+  # Ambiguous call resolved as CORE::select(), qualify as such or use &
+  return(CORE::select($vec, undef, undef, 0) > 0);
+}
+
 
 sub _data_sig_pipe_build {
   my( $self ) = @_;
   return unless USE_SIGNAL_PIPE;
   my $fake = 128;
-  
-  unless( %SIG2NUM ) {
-    foreach my $sig ( keys %_safe_signals ) {
-      my $n = eval "POSIX::SIG$sig()";
-      # warn $@;
-      if( $@ ) {    # AKA : RUNNING_IN_HELL
-          # The number used is less important then the fact that it has
-          # a unique number assigned to it
-          $n = $fake++;
-          _trap "<sg> SIG$sig not defined and $n > 255" if $n > 255;
-      }
-      else {
-          # paranoid check
-          _trap "<sg> SIG$sig is out of range ($n)" if $n > 127;
-      }
-      $SIG2NUM{ $sig } = $n;
-      $NUM2SIG{ $n } = $sig;
-    }
-    # warn join "\n", map { "$_: $SIG2NUM{$_}" } sort keys %SIG2NUM;
-    # we need CLD to be named CHLD
-    $SIG2NUM{ CLD } = $SIG2NUM{ CHLD };
-    $NUM2SIG{ $SIG2NUM{ CHLD } } = 'CHLD';
-    $NUM2SIG{ $SIG2NUM{ CLD } } = 'CHLD' if $SIG2NUM{ CLD };
-    # warn join "\n", map { "$_: $_safe_signals{$_}" } sort keys %_safe_signals;
-  }
 
   # Associate the pipe with this PID
   $signal_pipe_pid = $$;
@@ -713,14 +819,19 @@ sub _data_sig_pipe_build {
   # Mess with the signal mask
   $self->_data_sig_mask_all;
 
-  # Open the signal pipe
+  # Open the signal pipe.
+  # TODO - Normally POE::Pipe::OneWay will do the right thing.  Why
+  # are we overriding its per-platform autodetection?
   if (RUNNING_IN_HELL) {
     ( $signal_pipe_read, $signal_pipe_write ) = POE::Pipe::OneWay->new('inet');
   }
   else {
     ( $signal_pipe_read, $signal_pipe_write ) = POE::Pipe::OneWay->new('pipe');
   }
-  _trap "<sg> Error " . ($!+0) . " trying to create the signal pipe: $!" unless $signal_pipe_write;
+
+  unless ($signal_pipe_write) {
+    _trap "<sg> Error " . ($!+0) . " trying to create the signal pipe: $!";
+  }
 
   # Allows Resource::FileHandles to by-pass the queue
   $signal_pipe_read_fd = fileno $signal_pipe_read;
@@ -786,113 +897,136 @@ sub _data_sig_pipe_finalize {
 ### Send a signal "message" to the main thread
 ### Called from the top signal handlers
 sub _data_sig_pipe_send {
-  my $n = $SIG2NUM{ $_[1] };
-  if( ASSERT_DATA ) {
-    _trap "<sg> Unknown signal $_[1]" unless defined $n;
-  }
+  local $!;
+
+  my $signal_name = $_[1];
+
   if( TRACE_SIGNALS ) {
-    _warn "<sg> Caught SIG$_[1] ($n)";
+    _warn "<sg> Caught SIG$signal_name";
   }
-  
+
   return if $finalizing;
-  
+
   if( not defined $signal_pipe_pid ) {
-    _trap "<sg> $$ _data_sig_pipe_send called before signal pipe was initialized.";
+    _trap "<sg> _data_sig_pipe_send called before signal pipe was initialized.";
   }
+
+  # ugh- has_forked() can't be called fast enough.  This warning might
+  # show up before it is called.  Should we just detect forking and do it
+  # for the user?  Probably not...
+
   if( $$ != $signal_pipe_pid ) {
-    _trap "<sg> Kernel now running in a different process (is=$$ was=$signal_pipe_pid).  You must call call \$poe_kernel->has_forked in the child process.";
+    _warn(
+      "<sg> Signal caught in different process than POE::Kernel initialized " .
+      "(newPID=$$ oldPID=$signal_pipe_pid sig=$signal_name).\n"
+    );
+    _warn(
+      "Call POE::Kernel->has_forked() in the child process " .
+      "to relocate the signal handler.\n"
+    );
   }
 
+  # We're registering signals in a list.  Pipes have more finite
+  # capacity, so we'll just write a single-byte semaphore-like token.
+  # It's up to the reader to process the list.  Duplicates are
+  # permitted, and their ordering may be significant.  Precedent:
+  # http://search.cpan.org/perldoc?IPC%3A%3AMorseSignals
 
-  my $count = _data_sig_pipe_syswrite( pack( "C", $n ) );
-  if( ASSERT_DATA ) {
-    if( $count != 1 ) {
-      _trap "<sg> Wrote more than one byte (count=$count)";
+  push @pending_signals, [
+    $signal_name, # SIGINFO_NAME
+    $$,           # SIGINFO_SRC_PID
+  ];
+
+  if (TRACE_SIGNALS) {
+    _warn "<sg> Attempting signal pipe write";
+  }
+
+  my $count = syswrite( $signal_pipe_write, '!' );
+
+  # TODO - We need to crash gracefully if the write fails, but not if
+  # it's due to the pipe being full.  We might solve this by only
+  # writing on the edge of @pending_signals == 1 after the push().
+  # We assume @pending_signals > 1 means there's a byte in the pipe,
+  # so the reader will wake up to catch 'em all.
+
+  if ( ASSERT_DATA ) {
+    unless (defined $count and $count == 1) {
+      _trap "<sg> Signal pipe write failed: $!";
     }
   }
 }
 
-### write one signal number to the pipe
-sub _data_sig_pipe_syswrite {
-  my( $data ) = @_;
-  my $count = syswrite( $signal_pipe_write, $data );
-  if( defined $count and $count > 0 ) {
-    $! = 0;
-    if( TRACE_SIGNALS ) {
-      _warn "<sg> Wrote $count byte(s) to signal pipe";
-    }
-
-    return $count;
-  }
-
-  # if we got here, something bad happened
-  if( $! == EAGAIN or $! == EWOULDBLOCK ) {
-    _trap "<sg> Excessive signals detected; signal pipe full: $!";
-  }
-  _trap "<sg> Error " . ($!+0) . " writing to signal pipe: $!";
-}
-
 ### Read all signal numbers.
 ### Call the related bottom handler.  That is, inside the kernel loop.
 sub _data_sig_pipe_read {
   my( $self, $fileno, $mode ) = @_;
+
   if( ASSERT_DATA ) {
     _trap "Illegal mode=$mode on fileno=$fileno" unless
                                     $fileno == $signal_pipe_read_fd
                                 and $mode eq MODE_RD;
   }
-  my $data = $self->_data_sig_pipe_sysread();
-  return unless defined $data;
 
-  my $count = length $data;
-  if( TRACE_SIGNALS ) {
-    _warn "<sg> Read $count bytes from signal pipe";
+  # Read all data from the signal pipe.
+  # The data itself doesn't matter.
+  # TODO - If writes can happen on the edge of @pending_signals (from
+  # 0 to 1 element), then we oughtn't need to loop here.
+
+  while (1) {
+    my $octets_count = sysread( $signal_pipe_read, (my $data), 65536 );
+
+    next if $octets_count;
+    last if defined $octets_count;
+
+    last if $! == EAGAIN or $! == EWOULDBLOCK;
+
+    if (ASSERT_DATA) {
+      _trap "<sg> Error " . ($!+0) . " reading from signal pipe: $!";
+    }
+    elsif(TRACE_SIGNALS) {
+      _warn "<sg> Error " . ($!+0) . " reading from signal pipe: $!";
+    }
+
+    last;
   }
-  return unless $count;
 
-  for(my $q=0; $q< $count; $q++ ) {
-    my $n = unpack "C", substr( $data, $q, 1 );
-    next if $n == 0;
-    if( ASSERT_DATA ) {
-      _trap "Unknown signal number $n" unless $NUM2SIG{ $n };
+  # Double buffer signals.
+  # The intent is to avoid a race condition by processing the same
+  # buffer that new signals go into.
+
+  return unless @pending_signals;
+  my @signals = @pending_signals;
+  @pending_signals = ();
+
+  if (TRACE_SIGNALS) {
+    _warn "<sg> Read " . scalar(@signals) . " signals from the list";
+  }
+
+  foreach my $signal (@signals) {
+    my $signal_name    = $signal->[SIGINFO_NAME];
+    my $signal_src_pid = $signal->[SIGINFO_SRC_PID];
+
+    # Ignore signals from other processes.
+    # This can happen if we've fork()ed without calling has_forked()
+    # to reset the signals subsystem.
+    #
+    # TODO - We might be able to get rid of has_forked() if PID
+    # mismatches are detected.
+
+    next if $signal_src_pid != $$;
+
+    if( $signal_name eq 'CHLD' ) {
+      _loop_signal_handler_chld_bottom( $signal_name );
     }
-    my $sig = $NUM2SIG{ $n };
-    if( $sig eq 'CHLD' ) {
-      _loop_signal_handler_chld_bottom( $sig );
-    }
-    elsif( $sig eq 'PIPE' ) {
-      _loop_signal_handler_pipe_bottom( $sig );
+    elsif( $signal_name eq 'PIPE' ) {
+      _loop_signal_handler_pipe_bottom( $signal_name );
     }
     else {
-      _loop_signal_handler_generic_bottom( $sig );
+      _loop_signal_handler_generic_bottom( $signal_name );
     }
   }
 }
 
-### Read all signal numbers from the pipe
-sub _data_sig_pipe_sysread {
-  my $data = '';
-  # To avoid flooding the queue, we don't read the entire pipe at once
-  # PG- Mind you, I doubt signal flooding is ever going to be much of
-  # a problem, is it?
-  my $result = sysread( $signal_pipe_read, $data, 4096 ); # XXX
-  if( defined $result ) {
-    $! = 0;
-    return $data;
-  }
-  # Nonfatal sysread() error.  Return an empty list.
-  return '' if $! == EAGAIN or $! == EWOULDBLOCK;
-
-  if( ASSERT_DATA ) {
-    _trap "<sg> Error " . ($!+0) . " reading from signal pipe: $!";
-  }
-  elsif( TRACE_SIGNALS ) {
-    _warn "<sg> Error " . ($!+0) . " reading from signal pipe: $!";
-  }
-  return;
-}
-
-
 1;
 
 __END__
@@ -919,7 +1053,7 @@ handling.
 See L<POE::Kernel/Signal Watcher Methods> for POE's public signals
 API.
 
-See L<POE::Kernel/Resources> for for public information about POE
+See L<POE::Kernel/Resources> for public information about POE
 resources.
 
 See L<POE::Resource> for general discussion about resources and the
